from apispec_helper._internal_utils.post_init_base import PostInitBase


class NullValueAttributeRemover:
    def __init__(self, instance):
        self.__instance = instance

    def execute(self):
        null_value_attr_list = self._search_null_value_attributes()
        self._remove_null_value_from_attribute(null_value_attr_list)
        self._remove_null_value_from_dataclass_fields(null_value_attr_list)

    def _search_null_value_attributes(self):
        attribute_list = self.__instance.__dict__
        null_value_attr_list = list()

        for attribute in attribute_list:
            if getattr(self.__instance, attribute) is None:
                null_value_attr_list.append(attribute)

        return null_value_attr_list

    def _remove_null_value_from_attribute(self, null_value_attr_list):
        for null_value_attr in null_value_attr_list:
            delattr(self.__instance, null_value_attr)

    def _remove_null_value_from_dataclass_fields(self, null_value_attr_list):
        dataclass_fields = getattr(self.__instance, "__dataclass_fields__")

        for null_value_attr in null_value_attr_list:
            if null_value_attr in dataclass_fields.keys():
                dataclass_fields.pop(null_value_attr)

        setattr(self.__instance, "__dataclass_fields__", dataclass_fields)


class KeywordAttributeRenamer:
    def __init__(self, instance):
        self.__instance = instance
        self.__keyword_list = ["in_", "type_", "format_"]

    def execute(self):
        for attribute_name in self.__keyword_list:
            new_attribute_name = attribute_name.rstrip("_")
            self.__rename_attribute(old_attribute_name=attribute_name, new_attribute_name=new_attribute_name)
            self.__rename_dataclass_field(old_attribute_name=attribute_name, new_attribute_name=new_attribute_name)

    def __rename_attribute(self, old_attribute_name: str, new_attribute_name: str):
        if hasattr(self.__instance, old_attribute_name):
            attribute_value = getattr(self.__instance, old_attribute_name)
            setattr(self.__instance, new_attribute_name, attribute_value)

    def __rename_dataclass_field(self, old_attribute_name: str, new_attribute_name: str):
        if old_attribute_name in getattr(self.__instance, "__dataclass_fields__").keys():
            dataclass_fields = getattr(self.__instance, "__dataclass_fields__")
            dataclass_fields[new_attribute_name] = dataclass_fields.pop(old_attribute_name)
            dataclass_fields[new_attribute_name].name = new_attribute_name
            setattr(self.__instance, "__dataclass_fields__", dataclass_fields)


class DataclassHelperBase(PostInitBase):
    def __post_init__(self):
        super().__post_init__()
        NullValueAttributeRemover(self).execute()
        KeywordAttributeRenamer(self).execute()



from apispec_helper._internal_utils.dataclass_helper_base import DataclassHelperBase
from apispec_helper.component.content import ContentType
from dataclasses import dataclass


@dataclass
class RequestBody(DataclassHelperBase):
    content: ContentType

    description: str = None
    required: bool = None




from dataclasses import dataclass, field
from apispec_helper._internal_utils.dataclass_helper_base import DataclassHelperBase
from apispec_helper.basic_type.examples import ExamplesType
from typing import List, Dict
from typing import Union


@dataclass
class CommonTypeBase(DataclassHelperBase):
    type_: str
    description: str = None
    nullable: bool = None
    readOnly: bool = None
    writeOnly: bool = None
    enum: List = None

    example: dict = None
    examples: ExamplesType = None


@dataclass
class Object(DataclassHelperBase):
    type_: str = field(default="object", init=False)
    properties: Dict[str, CommonTypeBase]
    required: List[str] = None

    additionalProperties: bool | dict = None

    minProperties: int = None
    maxProperties: int = None

    example: dict = None
    examples: ExamplesType = None


@dataclass
class OneOf(DataclassHelperBase):
    oneOf: List[CommonTypeBase | str]
    example: dict = None
    examples: ExamplesType = None


class AnyOf:
    anyOf: dict | List[CommonTypeBase | Object] = {}
    nullable: bool = None
    description: str = None

    example: dict = None
    examples: ExamplesType = None


class IntegerFormat:
    INT32 = "int32"
    INT64 = "int64"


@dataclass
class Integer(CommonTypeBase):
    type_: str = field(default="integer", init=False)
    # refer to IntegerFormat
    format_: str = None

    minimum: int = None
    exclusiveMinimum: bool = None

    maximum: int = None
    exclusiveMaximum: bool = None

    multipleOf: int = None


class NumberFormat:
    FLOAT = "float"
    DOUBLE = "double"


@dataclass
class Number(CommonTypeBase):
    type_: str = field(default="number", init=False)
    # refer to NumberFormat
    format_: str = None

    minimum: int = None
    exclusiveMinimum: bool = None

    maximum: int = None
    exclusiveMaximum: bool = None

    multipleOf: int = None


class PreDefinedStringFormat:
    DATE = "date"
    DATETIME = "date-time"
    PASSWORD = "password"
    BYTE = "byte"
    BINARY = "binary"


@dataclass
class String(CommonTypeBase):
    type_: str = field(default="string", init=False)
    # refer to PreDefinedStringFormat
    format_: str = None
    pattern: str = None

    minLength: int = None
    maxLength: int = None


@dataclass
class Boolean(CommonTypeBase):
    type_: str = field(default="boolean", init=False)


@dataclass
class Null(CommonTypeBase):
    type_: str = field(default="null", init=False)


@dataclass
class Array(CommonTypeBase):
    """
    :param items: Check the "Mixed-Type Arrays" section in https://swagger.io/docs/specification/data-models/data-types
    """
    type_: str = field(default="array", init=False)

    minItems: int = None
    maxItems: int = None

    # dict for {}, empty dict
    # str for reference
    items: CommonTypeBase | dict | list | OneOf | str = None

    uniqueItems: bool = None


# str for reference
CommonType = Union[CommonTypeBase, str, Object, OneOf, AnyOf]



class ParameterStyle:
    FORM = "form"
    SIMPLE = "simple"
    LABEL = "label"
    MATRIX = "matrix"
    PIPE_DELIMITED = "pipeDelimited"
    SPACE_DELIMITED = "spaceDelimited"
    DEEP_OBJECT = "deepObject"



from apispec_helper.basic_type.common_type import Object, OneOf, AnyOf, Boolean, Null, Array
from apispec_helper.basic_type.common_type import PreDefinedStringFormat, String
from apispec_helper.basic_type.common_type import IntegerFormat, Integer
from apispec_helper.basic_type.common_type import NumberFormat, Number
from apispec_helper.basic_type.common_type import CommonType

from apispec_helper.basic_type.examples import Example, ExamplesType
from apispec_helper.basic_type.external_documentation import ExternalDocumentation
from apispec_helper.basic_type.server import ServerVariables, Server


__all__ = [
    "Object", "OneOf", "AnyOf", "Boolean", "Null", "Array",
    "PreDefinedStringFormat", "String",
    "IntegerFormat", "Integer",
    "NumberFormat", "Number",
    "CommonType",
    
    "Example", "ExamplesType",
    "ExternalDocumentation",
    "ServerVariables", "Server"
]


from dataclasses import dataclass
from apispec_helper._internal_utils.dataclass_helper_base import DataclassHelperBase
from apispec_helper._internal_utils.post_init_base import PostInitBase


class OAuthFlowFieldNotSpecifiedError(Exception):
    def __init__(self, flow_name, field_name):
        super().__init__(f"{field_name} is required for the OAuth flow {flow_name} but not specified")


class FlowFieldChecker:
    def __init__(self, instance, flow_name: str, field_name_list: list[str]):
        self.__instance = instance
        self.__flow_name = flow_name
        self.__field_name_list = field_name_list

    def execute(self):
        if hasattr(self.__instance, self.__flow_name):
            self.__check_field_exist()

    def __check_field_exist(self):
        for field_name in self.__field_name_list:
            if hasattr(getattr(self.__instance, self.__flow_name), field_name) is False:
                raise OAuthFlowFieldNotSpecifiedError(flow_name=self.__flow_name, field_name=field_name)


class OAuthFlowsDataclassHelperBase(PostInitBase):
    def __post_init__(self):
        super().__post_init__()
        FlowFieldChecker(instance=self, flow_name="implicit", field_name_list=["authorizationUrl", ]).execute()
        FlowFieldChecker(instance=self, flow_name="password", field_name_list=["tokenUrl", ]).execute()
        FlowFieldChecker(instance=self, flow_name="clientCredentials", field_name_list=["tokenUrl", ]).execute()
        FlowFieldChecker(instance=self, flow_name="authorizationCode", field_name_list=["tokenUrl", "authorizationUrl"]).execute()


@dataclass
class OAuthFlow(DataclassHelperBase):
    scopes: dict[str, str]

    refreshUrl: str = None

    # used for implicit / authorizationCode
    authorizationUrl: str = None

    # used for password / authorizationCode / clientCredentials
    tokenUrl: str = None


@dataclass
class OAuthFlows(DataclassHelperBase, OAuthFlowsDataclassHelperBase):
    implicit: OAuthFlow = None
    password: OAuthFlow = None
    clientCredentials: OAuthFlow = None
    authorizationCode: OAuthFlow = None


class SecuritySchemeType:
    API_KEY = "apiKey"
    HTTP = "http"
    OAUTH2 = "oauth2"
    OPENID_Connect = "openIdConnect"


class APIKeyLocation:
    QUERY = "query"
    HEADER = "header"
    COOKIE = "cookie"


class HTTPAuthenticationScheme:
    # refer to https://www.iana.org/asignments/http-authschemes/http-authschemes.xhtml
    BASIC = "basic"
    BEARER = "bearer"
    DIGEST = "digest"
    DPOP = "dpop"
    HOBA = "hoba"
    MUTUAL = "mutual"
    NEGOTIATE = "negotiate"
    OAUTH = "oauth"
    SCRAM_SHA_1 = "scram-sha-1"
    SCRAM_SHA_256 = "scram-sha-256"
    VAPID = "vapid"


@dataclass
class SecurityScheme(DataclassHelperBase):
    # refer to SecuritySchemeType
    type_: str

    description: str = None

    # used for API Key
    name: str = None

    # refer to APIKeyLocation
    in_: str = None

    # used for HTTP
    scheme: str = None
    # required when scheme is "bearer
    bearerFormat: str = None

    # used for oauth2
    flows: OAuthFlows = None

    # used for OpenID Connect
    openIdConnectUrl: str = None
